/**  File cairo_drv.c
 *
 * Driver using cairo library to make a PDF output
 */
#include <stdlib.h>
#include <string.h>

#include <gtk/gtk.h>
#include <cairo.h>
#include <cairo-ps.h>
#include <cairo-pdf.h>
#include <cairo-svg.h>

#include "device.h"
#include "draw.h"
#include "patterns.h"
#include "utils.h"

/*  Provisoire pour gg_initcmap */
#include "gg_gtkdrv.h"


#include "cairo_drv.h"

extern GdkColor gg_colors[MAXCOLORS];

extern char print_file[];

typedef struct {
  cairo_surface_t *surface;
  cairo_t         *cr;
  cairo_matrix_t  matrix;
  int    color;
  int    pattern;
  double linew;
  int    lines;
  int    linecap;
  int    linejoin;
  Page_geometry pg;
  double pixel_size;
  double dscale ,page_scalef ,page_heightf;
  double dheight ,dwidth;
  int bg;
} CairoDriver;
  
static CairoDriver *pcr = NULL; /* pointer on current context */
static CairoDriver hcr; /* harcopy context */

static void cr_init (CairoDriver *crd)
{
  crd->surface = NULL;
  crd->cr      = NULL;
  /* undefine all graphics state parameters */
  crd->color = -1;
  crd->pattern = -1;
  crd->linew = -1.0;
  crd->lines = -1;
  crd->linecap = -1;
  crd->linejoin = -1;
  /* device-independent routines */
  devupdatecmap    = NULL;
  devsetfrgbcolor  = cr_setfrgbcolor;
  devdrawpixel     = cr_drawpixel;
  devdrawpolyline  = cr_drawpolyline;
  devfillpolygon   = cr_fillpolygon;
  devdrawarc       = cr_drawarc;
  devfillarc       = cr_fillarc;
  devputpixmap     = cr_putpixmap;
  devputtext       = cr_puttext;
  /* devleavegraphics depends upon the driver */
  /* page geometry */
  crd->pg      	    = get_page_geometry();                 /* return page dimensions in pixels */
  crd->dheight 	    = crd->pg.height * 72.0 / crd->pg.dpi; /* height in points  (1 point == 1/72.0 inch) */
  crd->dwidth  	    = crd->pg.width  * 72.0 / crd->pg.dpi;
  crd->dscale       = (double) MIN2 (crd->pg.height ,crd->pg.width);
  crd->pixel_size   = 1.0 / crd->dscale;
  crd->page_scalef  = crd->dscale * 72.0 / crd->pg.dpi;
  crd->page_heightf = ((double)crd->pg.height) * 72.0 / crd->pg.dpi;
  /* background color */
  crd->bg = getbgcolor ();
}

/******************** Common to PDF driver and SVG driver ********************/

static void hcr_leavegraphics (void)
{
  cairo_stroke (hcr.cr);
  cairo_surface_finish  (hcr.surface);
  cairo_surface_destroy (hcr.surface);
  cairo_destroy (hcr.cr);
}


/******************** PostScript driver ********************/

static Device_entry dev_cairo_eps = {
  DEVICE_FILE,                              /* type */
  "EPScairo",                               /* *name : name of device */
  hcr_eps_initgraphics,                     /* function to initialize device */
  NULL,  //22  cr_eps_op_parser,            /* function to parse device-specific commands */
  NULL,  //22  cr_eps_gui_setup,            /* function (GUI interface) to setup device */
  "eps",                                    /* filename extension */
  FALSE, //  TRUE,                          /* device has its own fonts */
  TRUE,  // FALSE,		   	    /* font antialiasing */
   {3300, 2550, 300.0},                     /* device defaults */
  NULL				   	    /* device private data */
};

/* TODO load fonts in device */

int register_cairo_eps_drv (void)
{
  return register_device (dev_cairo_eps);
}

int hcr_eps_initgraphics (void)           /* initgraphics */
{
  cr_init (&hcr);   		
  devleavegraphics = hcr_leavegraphics;		 
  /* cairo_eps_surface_create () needs a filename here */
  hcr.surface = cairo_ps_surface_create (print_file ,hcr.dwidth ,hcr.dheight);
  cairo_ps_surface_set_eps (hcr.surface ,TRUE);
  hcr.cr      = cairo_create (hcr.surface);
  /*                               	 xx   yx  xy        yy      x0   y0       */
  cairo_matrix_init (&hcr.matrix ,hcr.page_scalef ,0. ,0. ,-hcr.page_scalef ,0. ,hcr.page_heightf);
  cairo_set_matrix  (hcr.cr ,&hcr.matrix);
  pcr = &hcr;
  return RETURN_SUCCESS;
}


/******************** PDF driver ********************/

static Device_entry dev_cairo_pdf = {
  DEVICE_FILE,                              /* type */
  "PDFcairo",                               /* *name : name of device */
  hcr_pdf_initgraphics,                      /* function to initialize device */
  NULL,  //22  cr_pdf_op_parser,            /* function to parse device-specific commands */
  NULL,  //22  cr_pdf_gui_setup,            /* function (GUI interface) to setup device */
  "pdf",                                    /* filename extension */
  FALSE, //  TRUE,                          /* device has its own fonts */
  TRUE,  // FALSE,		   	    /* font antialiasing */
   {3300, 2550, 300.0},                     /* device defaults */
  NULL				   	    /* device private data */
};

/* TODO load fonts in device */

int register_cairo_pdf_drv (void)
{
  return register_device (dev_cairo_pdf);
}

int hcr_pdf_initgraphics (void)           /* initgraphics */
{
  cr_init (&hcr);   		
  devleavegraphics = hcr_leavegraphics;		 
  /* cairo_pdf_surface_create () needs a filename here */
  hcr.surface = cairo_pdf_surface_create (print_file ,hcr.dwidth ,hcr.dheight);
  hcr.cr      = cairo_create (hcr.surface);
  /*                               	 xx   yx  xy        yy      x0   y0       */
  cairo_matrix_init (&hcr.matrix ,hcr.page_scalef ,0. ,0. ,-hcr.page_scalef ,0. ,hcr.page_heightf);
  cairo_set_matrix  (hcr.cr ,&hcr.matrix);
  pcr = &hcr;
  return RETURN_SUCCESS;
}

/******************** PNG driver ********************/

static Device_entry dev_cairo_png = {
  DEVICE_FILE,                              /* type */
  "PNGcairo",                               /* *name : name of device */
  hcr_png_initgraphics,                     /* function to initialize device */
  NULL,           			    /* function to parse device-specific commands */
  NULL,           			    /* function (GUI interface) to setup device */
  "png",                                    /* filename extension */
  FALSE,                          	    /* device has its own fonts */
  TRUE, 		   	    	    /* font antialiasing */
  {DEFAULT_PAGE_WIDTH, DEFAULT_PAGE_HEIGHT, 72.0},  /* device defaults */
  NULL				   	    /* device private data */
};

int register_cairo_png_drv (void)
{
  return register_device (dev_cairo_png);
}

/**
 * PNG needs cairo_surface_write_to_png()
 */
void hcr_png_leavegraphics (void)
{
  cairo_stroke (hcr.cr);
  cairo_surface_write_to_png (hcr.surface ,print_file);
  cairo_surface_finish  (hcr.surface);
  cairo_surface_destroy (hcr.surface);
  cairo_destroy (hcr.cr);
}

int hcr_png_initgraphics (void)           /* initgraphics */
{
  Page_geometry pg;
  cr_init (&hcr);   	
  devleavegraphics = hcr_png_leavegraphics;
  pg      	   = get_page_geometry();                 /* return page dimensions in pixels */
  hcr.surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32 ,(int)pg.width ,(int)pg.height);
  hcr.cr      = cairo_create (hcr.surface);
  /*                               xx  	   yx  xy   yy  	 x0   y0       */
  cairo_matrix_init (&hcr.matrix ,hcr.page_scalef ,0. ,0. ,-hcr.page_scalef ,0. ,hcr.page_heightf);
  cairo_set_matrix  (hcr.cr ,&hcr.matrix);
  pcr = &hcr;
  return RETURN_SUCCESS;
}


/******************** SVG driver ********************/

static Device_entry dev_cairo_svg = {
  DEVICE_FILE,                              /* type */
  "SVGcairo",                               /* *name : name of device */
  hcr_svg_initgraphics,                      /* function to initialize device */
  NULL,         			    /* function to parse device-specific commands */
  NULL,         			    /* function (GUI interface) to setup device */
  "svg",                                    /* filename extension */
  FALSE, //  TRUE,                          /* device has its own fonts */
  TRUE,  // FALSE,		   	    /* font antialiasing */
  {DEFAULT_PAGE_WIDTH, DEFAULT_PAGE_HEIGHT, 72.0},  /* device defaults */
  NULL				   	    /* device private data */
};

int register_cairo_svg_drv (void)
{
  return register_device (dev_cairo_svg);
}

int hcr_svg_initgraphics (void)           /* initgraphics */
{
  cr_init (&hcr); 
  devleavegraphics = hcr_leavegraphics;
  hcr.surface = cairo_svg_surface_create (print_file ,hcr.dwidth ,hcr.dheight);
  hcr.cr      = cairo_create (hcr.surface);
  /*                               xx  	   yx  xy   yy  	 x0   y0       */
  cairo_matrix_init (&hcr.matrix ,hcr.page_scalef ,0. ,0. ,-hcr.page_scalef ,0. ,hcr.page_heightf);
  cairo_set_matrix  (hcr.cr ,&hcr.matrix);
  pcr = &hcr;
  return RETURN_SUCCESS;
}


/******************** Functions common to all drivers ********************/

void cr_setpen (const Pen *pen)
{
  if (pen->color != hcr.color || pen->pattern != pcr->pattern) {
    fRGB *frgb  = get_frgb (pen->color);
    pcr->color 	= pen->color;
    pcr->pattern = pen->pattern;
    cairo_set_source_rgb (pcr->cr ,frgb->red ,frgb->green ,frgb->blue);  /* setcolor */
  }
}

/**
 *  Set foreground color in cmap_table[fg] directly
 */
void cr_setfrgbcolor (double fred ,double fgreen ,double fblue)
{
  CMap_entry ce ,*ce_old;
  int fg;
  cairo_set_source_rgb (pcr->cr ,fred ,fgreen ,fblue);
  fg = getcolor ();
  ce_old = get_cmap_entry (fg);
  if (ce_old != NULL) {
    fred   *= (MAXCOLORS - 1);
    fgreen *= (MAXCOLORS - 1);
    fblue  *= (MAXCOLORS - 1);
    ce.rgb.red   = (int) fred;
    ce.rgb.green = (int) fgreen;
    ce.rgb.blue  = (int) fblue;
    ce.cname     = ce_old->cname;
    ce.ctype     = COLOR_MAIN;
    ce.tstamp    = 0;
    store_color (fg ,ce);
  }
}


void cr_setdrawbrush (void)  // pdf_setdrawbrush
{
  int i ,ls;
  double *cdarr ,darr_len ,scale ,lw;
  Pen pen;
  pen = getpen ();
  cr_setpen  (&pen);
  lw = getlinewidth ();
  cairo_set_line_width (pcr->cr ,lw);
  ls = getlinestyle ();
  if (ls > 1) {
    darr_len = dash_array_length[ls];
    cdarr = xmalloc ( darr_len * SIZEOF_DOUBLE);
    scale = MAX2 (1., lw) / pcr->page_scalef;
    for (i = 0; i < darr_len; i++) {
      cdarr[i] = scale * (double)dash_array[ls][i];
    }
    cairo_set_dash (pcr->cr ,cdarr ,darr_len ,cdarr[0]);
  } else {
    /* dashing disabled */
    cairo_set_dash (pcr->cr ,NULL ,0 ,0.);
  }

//    int 	 lc = getlinecap   ();
//    int 	 lj = getlinejoin  ();
  // dash_array[style][i];

}

/* il semble que cette fonction n'est pas utilise */
void cr_drawpixel (VPoint vp)  // drawpixel
{
  printf ("cr_drawpixel : A FAIRE\n");
}

void cr_drawpolyline (VPoint *vps ,int n ,int mode)   // gg_drawpolyline
{
  int i;
  if (getlinestyle() == 0) return;
  cr_setdrawbrush ();
  //22  pdf_setlineprops();
  cairo_move_to (pcr->cr ,vps[0].x ,vps[0].y);
  for (i = 1; i < n; i++) {
    cairo_line_to (pcr->cr ,vps[i].x ,vps[i].y);
  }
  if (mode == POLYLINE_CLOSED) cairo_line_to (pcr->cr ,vps[0].x ,vps[0].y);
  cairo_save   	(pcr->cr);                          /* stack-pen-size-color */
  cairo_stroke 	(pcr->cr);
  cairo_restore (pcr->cr);                          /* stack-pen-size-color */
}


void cr_fillpolygon (VPoint *vps ,int n)
{
  int i ,j ,k ,l ,stride ,cindex;
  cairo_pattern_t * source;
  Pen pen;
  cairo_surface_t *patsurf;
  unsigned char data[16][16][4];

  pen = getpen ();
  if (pen.pattern == 0) return;
  cr_setpen (&pen);
  cairo_save (pcr->cr);                            /* stack-pen-size-color */
  if (getfillrule () == FILLRULE_WINDING) {
    cairo_set_fill_rule (pcr->cr ,CAIRO_FILL_RULE_WINDING);
  } else {
    cairo_set_fill_rule (pcr->cr ,CAIRO_FILL_RULE_EVEN_ODD);
  }
  cairo_move_to (pcr->cr ,vps[0].x ,vps[0].y);
  for (i = 1; i < n; i++) {
    cairo_line_to (pcr->cr ,vps[i].x ,vps[i].y);
  }
  if (pen.pattern > 1) {
    cairo_clip (pcr->cr);
    /* set scale to render pixmap at correct scale */
    cairo_identity_matrix (pcr->cr);
    stride  = cairo_format_stride_for_width (CAIRO_FORMAT_ARGB32 ,16);
    patsurf = cairo_image_surface_create_for_data ((unsigned char*) data ,CAIRO_FORMAT_ARGB32 ,16 ,16 ,stride);
    /* Convert XBM to CAIRO_FORMAT_ARGB32   */
    cindex = pen.color;
    for (j = 0; j < 256; j++) {
      k = j%16;
      l = j/16;
      if ((pat_bits[pen.pattern][j/8] >> (j%8)) & 0x01) {
	/* the bit is set */
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	data[l][k][0] = gg_colors[cindex].blue;
	data[l][k][1] = gg_colors[cindex].green;
	data[l][k][2] = gg_colors[cindex].red;
	data[l][k][3] = 0xff;
#else
	data[l][k][0] = 0xff;
	data[l][k][1] = gg_colors[cindex].red;
	data[l][k][2] = gg_colors[cindex].green;
	data[l][k][3] = gg_colors[cindex].blue;
#endif
      } else {
	/* transparent */
	data[l][k][0] = data[l][k][1] = data[l][k][2] = data[l][k][3] = 0x00;
      }
    }
    cairo_set_source_surface  (pcr->cr ,patsurf ,0. ,0.);
    source  = cairo_get_source (pcr->cr);
    cairo_pattern_set_extend (source ,CAIRO_EXTEND_REPEAT);
    cairo_paint (pcr->cr);
    cairo_restore (pcr->cr);                          /* stack-pen-size-color */
  } else {
    cairo_fill (pcr->cr);
    cairo_restore (pcr->cr);                          /* stack-pen-size-color */
  }
}


/**
 *
 */
void cr_drawarc (VPoint vp1 ,VPoint vp2 ,int angle1 ,int angle2)  // pdf_drawarc
{
  double xc ,yc ,rx ,ry ,a1 ,a2;
  cr_setdrawbrush ();
  cairo_save (pcr->cr);
  xc = (vp1.x + vp2.x) / 2;
  yc = (vp1.y + vp2.y) / 2;
  rx    = fabs (vp2.x - vp1.x) / 2;
  ry    = fabs (vp2.y - vp1.y) / 2;
  a1 = ((double)angle1) * M_PI / 180.0;
  a2 = ((double)angle2) * M_PI / 180.0;
  cairo_new_sub_path (pcr->cr);
  if (rx == ry) {
    /* cairo_arc() draws only circles */
    cairo_arc (pcr->cr ,xc ,yc ,rx ,a1 ,a2);
  } else {
    /* we change the scale along Ox */
    double ell = rx / ry;
  /*                               	 xx  	         yx  xy        yy    	 	  x0                     y0       */
    cairo_matrix_init (&pcr->matrix ,pcr->page_scalef*ell ,0. ,0. ,-pcr->page_scalef ,(1.-ell)*pcr->page_scalef*xc ,pcr->page_heightf);
    cairo_set_matrix  (pcr->cr ,&pcr->matrix);
    cairo_arc (pcr->cr ,xc ,yc ,ry ,a1 ,a2);
  }
  cairo_stroke  (pcr->cr);
  cairo_restore (pcr->cr);                          /* stack-pen-size-color */
}

void cr_fillarc(VPoint vp1 ,VPoint vp2 ,int angle1 ,int angle2 ,int mode)
{
  double xc ,yc ,rx ,ry ,a1 ,a2;
  int j ,k ,l ,stride ,cindex;
  cairo_pattern_t * source;
  Pen pen;
  cairo_surface_t *patsurf;
  unsigned char data[16][16][4];

  pen = getpen ();
  if (pen.pattern == 0) return;
  cr_setpen (&pen);
  cairo_save (pcr->cr);
  xc = (vp1.x + vp2.x) / 2;
  yc = (vp1.y + vp2.y) / 2;
  rx    = fabs (vp2.x - vp1.x) / 2;
  ry    = fabs (vp2.y - vp1.y) / 2;
  a1 = ((double)angle1) * M_PI / 180.0;
  a2 = ((double)angle2) * M_PI / 180.0;
  cairo_new_sub_path (pcr->cr);
  if (mode == ARCFILL_PIESLICE) {
    cairo_move_to (pcr->cr ,xc ,yc);
  }
  if (rx == ry) {
    /* cairo_arc() draws only circles */
    cairo_arc (pcr->cr ,xc ,yc ,rx ,a1 ,a2);
  } else {
    /* we change the scale along Ox */
    double ell = rx / ry;
    /*                               xx  	         yx  xy   yy    	 	 x0                     y0       */
    cairo_matrix_init (&pcr->matrix ,pcr->page_scalef*ell ,0. ,0. ,-pcr->page_scalef ,(1.-ell)*pcr->page_scalef*xc ,pcr->page_heightf);
    cairo_set_matrix  (pcr->cr ,&pcr->matrix);
    cairo_arc (pcr->cr ,xc ,yc ,ry ,a1 ,a2);
  }
  if (mode == ARCFILL_PIESLICE) {
    cairo_close_path (pcr->cr);
  }
  if (pen.pattern > 1) {
    cairo_clip (pcr->cr);
    /* set scale to render pixmap at correct scale */
    cairo_identity_matrix (pcr->cr);
    stride  = cairo_format_stride_for_width (CAIRO_FORMAT_ARGB32 ,16);
    patsurf = cairo_image_surface_create_for_data ((unsigned char*) data ,CAIRO_FORMAT_ARGB32 ,16 ,16 ,stride);
    /* Convert XBM to CAIRO_FORMAT_ARGB32   */
    cindex = pen.color;
    for (j = 0; j < 256; j++) {
      k = j%16;
      l = j/16;
      if ((pat_bits[pen.pattern][j/8] >> (j%8)) & 0x01) {
	/* the bit is set */
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	data[l][k][0] = gg_colors[cindex].blue;
	data[l][k][1] = gg_colors[cindex].green;
	data[l][k][2] = gg_colors[cindex].red;
	data[l][k][3] = 0xff;
#else
	data[l][k][0] = 0xff;
	data[l][k][1] = gg_colors[cindex].red;
	data[l][k][2] = gg_colors[cindex].green;
	data[l][k][3] = gg_colors[cindex].blue;
#endif
      } else {
	/* transparent */
	data[l][k][0] = data[l][k][1] = data[l][k][2] = data[l][k][3] = 0x00;
      }
    }
    cairo_set_source_surface  (pcr->cr ,patsurf ,0. ,0.);
    source  = cairo_get_source (pcr->cr);
    cairo_pattern_set_extend (source ,CAIRO_EXTEND_REPEAT);
    cairo_paint (pcr->cr);
    cairo_restore (pcr->cr);                          /* stack-pen-size-color */
  } else {
    cairo_fill (pcr->cr);
    cairo_restore (pcr->cr);                          /* stack-pen-size-color */
  }
}

int cr_op_parser(char *opstring)
{
        return RETURN_SUCCESS;
}

/* TODO: transparent pixmaps */

/**
 * Used by t1fonts.c to draw glyphs via devputpixmap
 */
void cr_putpixmap (VPoint vp ,int width ,int height ,char *databits ,   // devputpixmap
			  int pixmap_bpp ,int bitmap_pad ,int pixmap_type)
{
  unsigned char *data;
  int cindex;
  int i ,j, k ,stride;
  double x ,y ,pixscale;
  cairo_surface_t  *s;
  Page_geometry    pg;
  cairo_save (pcr->cr);
  pg = get_page_geometry ();                 /* return page dimensions in pixels */
  pixscale = 72. / pg.dpi;                   /* scaling if page resolution differs from 72 dpi */
  cairo_matrix_init (&pcr->matrix ,pixscale ,0. ,0. ,pixscale ,0. ,0.0);
  cairo_set_matrix  (pcr->cr ,&pcr->matrix);
  data   = xmalloc (width * height * 4);
  stride = cairo_format_stride_for_width (CAIRO_FORMAT_ARGB32 ,width);
  s      = cairo_image_surface_create_for_data (data ,CAIRO_FORMAT_ARGB32 ,width ,height ,stride);
  i = 0;
  for (k = 0; k < height; k++) {
    for (j = 0; j < width; j++) {
      cindex = (unsigned char) (databits)[k*width+j];
      if (cindex != pcr->bg) {
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	data[i] = gg_colors[cindex].blue;  i++;
	data[i] = gg_colors[cindex].green; i++;
	data[i] = gg_colors[cindex].red;   i++;
#else
	data[i] = gg_colors[cindex].red;   i++;
	data[i] = gg_colors[cindex].green; i++;
	data[i] = gg_colors[cindex].blue;  i++;
#endif
	data[i] = 0xff;                    i++;  /* opaque */
      } else {
	data[i] = 0; i++;
	data[i] = 0; i++;
	data[i] = 0; i++;
	data[i] = 0; i++;                        /* transparent */
      }
    }
  }
   x = pcr->page_scalef  * vp.x      		 / pixscale;
   y = (pcr->page_heightf  -vp.y* pcr->page_scalef) / pixscale;
   cairo_set_source_surface (pcr->cr ,s ,x ,y);
   cairo_paint (pcr->cr);
  /* return to initial scaling. It seems that freeing data and s is done by cairo_restore */
  cairo_matrix_init (&pcr->matrix ,pcr->page_scalef ,0. ,0. ,-pcr->page_scalef ,0. ,pcr->page_heightf);
  cairo_set_matrix  (pcr->cr ,&pcr->matrix);
  cairo_restore (pcr->cr);
}

void cr_puttext (VPoint vp ,char *s ,int len ,int font,
     TextMatrix *tm ,int underline ,int overline ,int kerning)
{
  fprintf (stderr ,"cr_puttext not yet implemented\n");
}




